/*******************************************************************************
 * Copyright 2015 - 2018 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.mahoe.assistant;

import java.rmi.RemoteException;

import de.frankmuenster.mahoe.MahoeSynchronizeBackend;
import de.frankmuenster.mahoe.utils.I18N;
import de.willuhn.jameica.hbci.HBCI;
import de.willuhn.jameica.hbci.SynchronizeOptions;
import de.willuhn.jameica.hbci.rmi.HBCIDBService;
import de.willuhn.jameica.hbci.rmi.Konto;
import de.willuhn.jameica.system.Application;
import de.willuhn.jameica.system.OperationCanceledException;
import de.willuhn.logging.Logger;
import de.willuhn.util.ApplicationException;

/**
 * <p>
 * Abstract class to create an account. Show the dialog, creates a new account
 * and set default/basic values. Then calls the specific setup routine, which
 * must be implemented by all sub classes.
 * </p>
 * <p>
 * The abstract class handles the following properties form the
 * <code>CreateInfo</code> class object:
 * <ul>
 * <li>Account type</li>
 * <li>Comment</li>
 * <li>Description</li>
 * <li>Owner</li>
 * <li>Save account sheets</li>
 * <li>Save folder</li>
 * </ul>
 * </p>
 *
 * @author Frank Mnster
 * @param <I>
 *            the bank specific <code>CreateInfo</code> class
 * @param <K>
 *            the bank specific <code>AbstractCreateAccountDialog</code> class
 */
public abstract class AbstractCreateAccountAssistant<I extends CreateInfo, K extends AbstractCreateAccountDialog<I>>
        implements CreateAccountAssistant {

    /**
     * Standardkonstruktor
     */
    public AbstractCreateAccountAssistant() {

        // Absichtlich leer
    }

    /**
     * @see de.frankmuenster.mahoe.assistant.CreateAccountAssistant#createAccount()
     */
    @Override
    public Konto createAccount() throws CreateAccountException, OperationCanceledException {

        I info = null;
        try {
            final K infoDialog = this.getDialog();
            info = infoDialog.open();
        } catch (final OperationCanceledException e) { // NOPMD
            throw e;
        } catch (final Exception e) { // NOPMD
            Logger.warn(I18N.tr("CreateAccountAssistant.openDialogFailedLog", //$NON-NLS-1$
                    e.getMessage()));
            throw new CreateAccountException(I18N.tr("CreateAccountAssistant.openDialogFailed"), e); //$NON-NLS-1$
        }
        return this.createKonto(info);
    }

    /**
     * @return the account specific dialog
     * @throws CreateAccountException
     */
    public abstract K getDialog() throws CreateAccountException;

    /**
     * Set the bank specific account parameters. At least <code>kto</code>,
     * <code>blz</code> and <code>kundenkennung</code>.
     *
     * @param konto
     *            the account to add KTO and BLZ
     * @param info
     *            the account information for the new account
     * @throws RemoteException
     */
    protected abstract void setBankSpecificAccountParameter(Konto konto, I info) throws RemoteException;

    /**
     * Set the bank specific parameters. The subclass implementation needs to
     * take care about storing added/updated values.
     *
     * @param info
     *            the info object from the dialog
     * @param konto
     *            the account created
     * @throws RemoteException
     * @throws ApplicationException
     */
    protected abstract void setBankSpecificMetaData(I info, Konto konto) throws RemoteException, ApplicationException;

    /**
     * Creates the account with the parameters of the give info object. Calls
     * bank specific implementations to set <code>kto</code>, <code>blz</code>
     * and <code>kundenkennung</code> and other account parameter and META-data.
     *
     * @param info
     *            the info object holding the parameters of the new account
     * @return the new account
     * @throws CreateAccountException
     *             if something went wrong during account creation
     */
    private Konto createKonto(final I info) throws CreateAccountException {

        HBCIDBService dbService = null;
        try {
            dbService = (HBCIDBService) Application.getServiceFactory().lookup(HBCI.class, "database"); //$NON-NLS-1$
        } catch (final Exception e) { // NOPMD
            Logger.error(I18N.tr("CreateAccountAssistant.noDbAccess"), e); //$NON-NLS-1$
            throw new CreateAccountException(I18N.tr("CreateAccountAssistant.noDbAccess"), //$NON-NLS-1$
                    e);
        }
        Konto konto;
        try {
            konto = dbService.createObject(Konto.class, null);
            // set common fields from CreateInfo object
            konto.setName(info.getOwner());
            konto.setBezeichnung(info.getDescription());

            konto.setKommentar(info.getComment());
            konto.setAccountType(info.getAccountType().getValue());

            konto.setFlags(Konto.FLAG_NONE);
            konto.setBackendClass(MahoeSynchronizeBackend.BACKEND_NAME);

            // Set bank specific parameter, at least kto, blz and kundenkennung
            this.setBankSpecificAccountParameter(konto, info);

            konto.store();

            // Sync-Options setzen
            final SynchronizeOptions sync = new SynchronizeOptions(konto);
            sync.setAutoSaldo(false);
            sync.setSyncAuslandsUeberweisungen(false);
            sync.setSyncKontoauszuege(true);
            sync.setSyncOffline(false);
            sync.setSyncSaldo(true);
            sync.setSyncSepaDauerauftraege(false);
            sync.setSyncSepaLastschriften(false);

            // Meta data

            // Account specific
            this.setBankSpecificMetaData(info, konto);

        } catch (RemoteException | ApplicationException e) {

            Logger.error(I18N.tr("CreateAccountAssistant.createAccountFailedLog", //$NON-NLS-1$
                    e.getMessage()), e);
            throw new CreateAccountException(I18N.tr("CreateAccountAssistant.createAccountFailed"), //$NON-NLS-1$
                    e);
        }

        return konto;

    }
}
