/*******************************************************************************
 * Copyright 2015 - 2019 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.mahoe.assistant;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.rmi.RemoteException;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import de.frankmuenster.mahoe.MahoeSynchronizeBackend;
import de.frankmuenster.mahoe.Settings;
import de.frankmuenster.mahoe.banks.aktien.assistant.AktienAccountMigrator;
import de.frankmuenster.mahoe.banks.bos.BankOfScotland;
import de.frankmuenster.mahoe.banks.bsh.SchwaebischHall;
import de.frankmuenster.mahoe.banks.bsh.assistant.BshAccountMigrator;
import de.frankmuenster.mahoe.banks.santander.assistant.SantanderAccountMigrator;
import de.frankmuenster.mahoe.banks.targobank.assistant.TargoBankAccountMigrator;
import de.frankmuenster.mahoe.utils.I18N;
import de.frankmuenster.mahoe.utils.Utils;
import de.willuhn.datasource.rmi.DBIterator;
import de.willuhn.jameica.hbci.rmi.HBCIDBService;
import de.willuhn.jameica.hbci.rmi.Konto;
import de.willuhn.jameica.messaging.Message;
import de.willuhn.jameica.messaging.MessageConsumer;
import de.willuhn.jameica.messaging.StatusBarMessage;
import de.willuhn.jameica.messaging.SystemMessage;
import de.willuhn.jameica.plugin.Plugin;
import de.willuhn.jameica.system.Application;
import de.willuhn.logging.Logger;
import de.willuhn.util.ApplicationException;

/**
 * Started den Account-Migrator erst nach vollstndigem Systemstart
 *
 * @author Frank Mnster
 */
public class MahoeUpdateHandler implements MessageConsumer {

    /**
     * Liste mit allen bekannten AccountMigratoren/Updatern
     */
    private final transient Set<Class<? extends AccountMigrator>> accountMigrators = new HashSet<>();

    /**
     * Erstellt die Migrator-Klasse und registriert sich selbst in der MessagingFactory
     */
    public MahoeUpdateHandler() {

        this.accountMigrators.add(AktienAccountMigrator.class);
        this.accountMigrators.add(BshAccountMigrator.class);
        this.accountMigrators.add(SantanderAccountMigrator.class);
        this.accountMigrators.add(TargoBankAccountMigrator.class);
    }

    /**
     * @see de.willuhn.jameica.messaging.MessageConsumer#autoRegister()
     */
    @Override
    public boolean autoRegister() {

        return true;
    }

    /**
     * @see de.willuhn.jameica.messaging.MessageConsumer#getExpectedMessageTypes()
     */
    @SuppressWarnings("rawtypes")
    @Override
    public Class[] getExpectedMessageTypes() {

        return new Class[] { SystemMessage.class };
    }

    /**
     * @throws ApplicationException
     * @see de.willuhn.jameica.messaging.MessageConsumer#handleMessage(de.willuhn.jameica.messaging.Message)
     */
    @Override
    public void handleMessage(final Message message) throws ApplicationException {

        final SystemMessage sysMsg = (SystemMessage) message;
        if (sysMsg.getStatusCode() == SystemMessage.SYSTEM_STARTED) {

            Logger.debug("System gestartet. Nun kann der Migrator gestartet werden."); //$NON-NLS-1$
            this.checkOldPlugins();
            // Update accounts to new version if necessary
            this.updateAccounts();
            // Update account to new MahoeBack end class
            this.pluginUpdate();
            // Check for BOS account and Selenium plugin
            this.checkBosAccount();
            // Unregister as message consumer
            this.unregister();
        }
    }

    private void checkBosAccount() {

        final boolean existsBosAcc = this.existsBosAccount();

        if (existsBosAcc) {
            final boolean seleniumInstalled = this.isSeleniumInstalled();

            if (!seleniumInstalled) {
                final StatusBarMessage message = new StatusBarMessage(
                        I18N.tr("MahoeUpdateHandler.seleniumPluginNeeded"), //$NON-NLS-1$
                        StatusBarMessage.TYPE_ERROR);
                Application.getMessagingFactory().sendMessage(message);
            }
        }

    }

    /**
     * Checks if the Selenium-Plugin is installed
     *
     * @return <i>true</i> in case the Selenium-Plugin is installed, else <i>false</i>.
     */
    private boolean isSeleniumInstalled() {

        final List<Plugin> plugins = Application.getPluginLoader().getInstalledPlugins();
        for (final Plugin plugin : plugins) {
            if ("hibiscus.scripting.selenium".equals(plugin.getManifest().getName())) { //$NON-NLS-1$
                Settings.setSeleniumInstalled(true);
                break;
            }
        }
        return Settings.isSeleniumInstalled();
    }

    /**
     * Checks if a BOS account exists based on the BLZ
     *
     * @return <i>true</i> in case a BOS account exists, else <i>false</i>.
     */
    private boolean existsBosAccount() {

        boolean existsBosAcc = false;
        HBCIDBService hcbiService;
        try {
            hcbiService = de.willuhn.jameica.hbci.Settings.getDBService();

            final DBIterator<Konto> accounts = hcbiService.createList(Konto.class);
            while (accounts.hasNext()) {
                final Konto account = accounts.next();
                if (BankOfScotland.BLZ.equals(account.getBLZ())) {
                    existsBosAcc = true;
                    break;
                }
            }
        } catch (final RemoteException e) {
            Logger.error(I18N.tr("MahoeUpdateHandler.errorLoopingAccounts"), e); //$NON-NLS-1$
        }

        return existsBosAcc;
    }

    /**
     * @throws ApplicationException
     */
    private void checkOldPlugins() throws ApplicationException {

        final List<Plugin> plugins = Application.getPluginLoader().getInstalledPlugins();

        for (final Class<? extends AccountMigrator> clazz : this.accountMigrators) {

            try {
                final Constructor<? extends AccountMigrator> cons = clazz.getConstructor();
                final AccountMigrator accountMigrator = cons.newInstance();
                for (final Plugin plugin : plugins) {
                    accountMigrator.migrate(plugin);
                }
            } catch (NoSuchMethodException | SecurityException | InstantiationException | IllegalAccessException
                    | IllegalArgumentException | InvocationTargetException e) {
                Logger.error(I18N.tr("MahoeUpdateHandler.checkforOldScriptingModulesFailed", e.getMessage()), e); //$NON-NLS-1$
            }
        }

    }

    /**
     * Perform updates on plugin level
     */
    private void pluginUpdate() {

        if (!Settings.isBackEndClassUpdated()) {
            // Update data provider
            this.updateToNewBackenClass();
            Settings.setBackEndClassUpdated(true);
        }
    }

    private void unregister() {

        // Unregister
        Application.getMessagingFactory().unRegisterMessageConsumer(this);
    }

    private void updateAccounts() throws ApplicationException {

        for (final Class<? extends AccountMigrator> clazz : this.accountMigrators) {

            try {
                final Constructor<? extends AccountMigrator> cons = clazz.getConstructor();
                final AccountMigrator accountMigrator = cons.newInstance();

                final HBCIDBService hcbiService = de.willuhn.jameica.hbci.Settings.getDBService();
                final DBIterator<Konto> accounts = hcbiService.createList(Konto.class);
                while (accounts.hasNext()) {
                    final Konto account = accounts.next();
                    accountMigrator.update(account);
                }
            } catch (NoSuchMethodException | SecurityException | InstantiationException | IllegalAccessException
                    | IllegalArgumentException | InvocationTargetException | RemoteException e) {
                Logger.error(I18N.tr("MahoeUpdateHandler.updateAccountFailed", e.getMessage()), e); //$NON-NLS-1$
            }
        }

    }

    /**
     * Updates backend class to new class name {@link MahoeSynchronizeBackend#BACKEND_NAME}
     */
    private void updateToNewBackenClass() {

        try {
            final HBCIDBService hcbiService = de.willuhn.jameica.hbci.Settings.getDBService();
            final DBIterator<Konto> accounts = hcbiService.createList(Konto.class);
            while (accounts.hasNext()) {
                final Konto account = accounts.next();
                final String backend = account.getBackendClass();
                Logger.debug("Backend class name: " + backend); //$NON-NLS-1$
                if (backend != null && backend.equals("de.frankmuenster.hibiscus.mahoe.MahoeSynchronizeBackend")) { //$NON-NLS-1$
                    account.setBackendClass(MahoeSynchronizeBackend.BACKEND_NAME);
                    account.store();
                    Logger.info(I18N.tr("MahoeUpdateHandler.updateToNewBackendClass", account.getBezeichnung())); //$NON-NLS-1$
                }
                if (SchwaebischHall.BLZ.equals(account.getBLZ())) {
                    account.setMeta(MahoeSynchronizeBackend.PW_ASK_ALWAYS, Utils.TEXT_NEIN);
                }
            }
        } catch (ApplicationException | SecurityException | IllegalArgumentException | RemoteException e) {
            Logger.error(I18N.tr("MahoeUpdateHandler.updateToNewBackendClassFailed", e.getMessage()), e); //$NON-NLS-1$
        }
    }
}
