/*******************************************************************************
 * Copyright 2015 - 2020 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.mahoe.banks.bcee.web;

import java.nio.charset.Charset;
import java.rmi.RemoteException;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.List;
import java.util.Locale;

import org.apache.commons.lang.StringUtils;

import com.gargoylesoftware.htmlunit.TopLevelWindow;
import com.gargoylesoftware.htmlunit.WebClient;
import com.gargoylesoftware.htmlunit.WebResponse;
import com.gargoylesoftware.htmlunit.html.HtmlPage;

import de.frankmuenster.mahoe.utils.I18N;
import de.frankmuenster.mahoe.utils.Utils;
import de.frankmuenster.mahoe.web.WebSiteException;
import de.frankmuenster.mahoe.web.WebSiteExceptionTypes;
import de.willuhn.jameica.hbci.Settings;
import de.willuhn.jameica.hbci.rmi.HBCIDBService;
import de.willuhn.jameica.hbci.rmi.Umsatz;
import de.willuhn.logging.Logger;
import de.willuhn.util.ApplicationException;

/**
 * Statische Implementierungen der Umsatzfunktionen
 *
 * @author Frank Mnster
 */
@SuppressWarnings("PMD.ExcessiveImports")
public final class BceeWebSiteTransactionsImpl {

    /**
     * I18N key for get transaction type failed message.
     */
    private static final String I18N_GET_TRANSACTION_TYPE_FAILED = // NOPMD
            "BceeWebSiteTransactionsImpl.getTransactionTypeFailed"; //$NON-NLS-1$

    /**
     * Anzahl der Felder in der CSV-Datei
     */
    private static final int CSV_FIELDS = 15;

    /**
     * Interne Liste mit allen Umstzen
     */
    private List<Umsatz> allTransactions;

    /**
     * Privater Konstruktor
     */
    public BceeWebSiteTransactionsImpl() {

        this.allTransactions = new ArrayList<>();
    }

    /**
     * Lscht die interen Umsatzliste
     */
    public void clear() {

        this.allTransactions.clear();
    }

    /**
     * Liste die gebuchten Umstze ein.
     *
     * @param webClient
     *            der angemeldete WebClient
     * @param accountInfo
     *            die Kontoinformationen
     * @return ein Liste mit den gebuchtenUmstzen
     * @throws WebSiteException
     */
    public List<Umsatz> getBookedTransactions(final WebClient webClient, // NOPMD
            final BceeWebAccountInformation accountInfo) throws WebSiteException {

        if (this.allTransactions.isEmpty()) {
            this.getTransactions(webClient, accountInfo);
        }

        final List<Umsatz> transactions = new ArrayList<>();
        for (final Umsatz umsatz : this.allTransactions) {

            try {
                if (StringUtils.isNotEmpty(umsatz.getKommentar())) {
                    transactions.add(umsatz);
                }
            } catch (final RemoteException e) {
                Logger.error(I18N.tr(BceeWebSiteTransactionsImpl.I18N_GET_TRANSACTION_TYPE_FAILED), e);
                throw new WebSiteException(I18N.tr(BceeWebSiteTransactionsImpl.I18N_GET_TRANSACTION_TYPE_FAILED), e);
            }

        }

        return transactions;
    }

    /**
     * Liest alle vorgemerkten Ums&auml;tze. Als Kriterium wird das Auszug
     * genommen. Fehlt dieser, wird der Umsatz als vorgemerkt markiert.
     *
     * @param webClient
     * @param accountInfo
     * @return List mit vorgemerkten Ums&auml;tzen
     * @throws WebSiteException
     * @throws ApplicationException
     */
    public List<Umsatz> getPreBookTransactions(final WebClient webClient, // NOPMD
            final BceeWebAccountInformation accountInfo) throws WebSiteException, ApplicationException {

        if (this.allTransactions.isEmpty()) {
            this.getTransactions(webClient, accountInfo);
        }

        final List<Umsatz> transactions = new ArrayList<>();
        for (final Umsatz umsatz : this.allTransactions) {

            try {
                if (StringUtils.isEmpty(umsatz.getKommentar())) {
                    umsatz.setFlags(Umsatz.FLAG_NOTBOOKED);
                    transactions.add(umsatz);
                }
            } catch (final RemoteException e) {
                Logger.error(I18N.tr //
                (BceeWebSiteTransactionsImpl.I18N_GET_TRANSACTION_TYPE_FAILED), e); // NOPMD
                throw new WebSiteException(I18N.tr(BceeWebSiteTransactionsImpl.I18N_GET_TRANSACTION_TYPE_FAILED), e);
            }
        }
        return transactions;
    }

    /**
     * Fgt die Salden hinzu und dreht die Reihenfolge um, damit inHibisucs die
     * Umsatz ID richtig ist (neue haben hhere IDs).
     *
     * @param transactions
     * @param accountInfo
     * @return Liste mit den gefundenen Ums&auml;tzen
     */
    private List<Umsatz> addSalden(final List<Umsatz> transactions, final BceeWebAccountInformation accountInfo) {

        // Teste auf Webseitensaldo und Saldo des ersten Umsatzes
        double saldo = accountInfo.getSaldo().doubleValue();

        try {
            for (final Umsatz umsatz : transactions) {
                umsatz.setSaldo(saldo);
                saldo = Utils.roundNumber(saldo - umsatz.getBetrag(), 2);
            }
        } catch (final RemoteException e) {
            Logger.error(I18N.tr("BceeWebSiteTransactionsImpl.setSaldoFailed"), e); //$NON-NLS-1$
        }

        // Liste umkehren
        final List<Umsatz> transWithSaldo = new ArrayList<>();
        for (int i = transactions.size() - 1; i >= 0; i--) {
            final Umsatz transacUmsatz = transactions.get(i);
            transWithSaldo.add(transacUmsatz);
        }

        return transWithSaldo;

    }

    /**
     * Liest alle Ums&auml;tze in die interen Liste
     *
     * @param webClient
     * @param accountInfo
     * @throws WebSiteException
     */
    private void getTransactions(final WebClient webClient, final BceeWebAccountInformation accountInfo)
            throws WebSiteException {

        final SimpleDateFormat sdfMonthName = new SimpleDateFormat("MMMM yyyy", //$NON-NLS-1$
                Locale.GERMAN);
        // final SimpleDateFormat sdfMonthNumber = new
        // SimpleDateFormat("01/MM/yyyy 00:00:00", Locale.GERMAN); //$NON-NLS-1$
        final SimpleDateFormat sdfQueryMonth = new SimpleDateFormat("01.MM.yyyy", Locale.GERMAN); //$NON-NLS-1$

        final List<Umsatz> transactions = new ArrayList<>();

        final Calendar lastMonth = Calendar.getInstance();
        lastMonth.setTime(accountInfo.getLastTransactionDate());
        lastMonth.add(Calendar.MONTH, -1);
        lastMonth.set(Calendar.DAY_OF_MONTH, 1);
        Logger.info(I18N.tr("BceeWebSiteTransactionsImpl.loadTransactionsfrom") //$NON-NLS-1$
                + sdfMonthName.format(lastMonth.getTime()));

        final Calendar month = Calendar.getInstance();
        month.set(Calendar.DAY_OF_MONTH, month.getActualMaximum(Calendar.DAY_OF_MONTH));
        month.set(Calendar.HOUR, 0);
        month.set(Calendar.MINUTE, 0);

        HtmlPage page;

        while (lastMonth.before(month)) {

            page = BceeWebSiteUtils.openAccount(webClient, accountInfo);

            Logger.debug("CSV anfragen fr Monat " + sdfQueryMonth.format(month.getTime())); //$NON-NLS-1$
            String javaScriptCode = "accountMovements.showAlertExport('" + sdfQueryMonth.format(month.getTime()) //$NON-NLS-1$
                    + "');"; //$NON-NLS-1$

            Logger.debug(webClient.getWebWindows().toString());

            page.executeJavaScript(javaScriptCode);
            final HtmlPage jsPage = (HtmlPage) webClient.getCurrentWindow().getEnclosedPage();

            Logger.debug("CSV Anfragen bestttigen"); //$NON-NLS-1$
            javaScriptCode = "accountMovements.confirmAlertExport();"; //$NON-NLS-1$
            jsPage.executeJavaScript(javaScriptCode);
            final WebResponse response = webClient.getCurrentWindow().getEnclosedPage().getWebResponse();

            if (response.getResponseHeaderValue("Content-Type").equals( //$NON-NLS-1$
                    "text/csv")) { //$NON-NLS-1$
                // Bank-Antwort is UFT-16LE (Little Endian) kodiert
                final String csv = response.getContentAsString(Charset.forName("UTF-16LE")); //$NON-NLS-1$
                try {
                    this.loadCsv(csv, transactions);
                } catch (final RemoteException e) {
                    Logger.error(I18N.tr("BceeWebSiteTransactionsImpl.errorReadingCSVForMonth", //$NON-NLS-1$
                            sdfQueryMonth.format(month.getTime())), e);
                }

            } else {
                Logger.debug(response.getContentAsString());
                throw new WebSiteException(I18N.tr("BceeWebSiteTransactionsImpl.exportCsvFailed"), //$NON-NLS-1$
                        WebSiteExceptionTypes.PAGE_NOT_FOUND);
            }

            // Untige TopLevelWindows schlieen
            final List<TopLevelWindow> webWindows = webClient.getTopLevelWindows();
            for (int i = 1; i < webWindows.size(); i++) {
                webWindows.get(i).close();
            }

            // Nchster Monat
            month.add(Calendar.MONTH, -1);

        }

        // Salden hinzufgen und Umstze ordnen
        this.allTransactions = this.addSalden(transactions, accountInfo);

    }

    /**
     * L&auml;d die Ums&auml;tze aus der CSV-Datei in die Umsatzliste.
     *
     * @param csv
     * @param transactions
     * @throws RemoteException
     */
    private void loadCsv(final String csv, final List<Umsatz> transactions) throws RemoteException {

        final String[] lines = csv.split("\r\n"); //$NON-NLS-1$

        for (int i = lines.length - 1; i >= 0; i--) {
            final String line = lines[i];
            final String[] transaction = line.split(";", -1); //$NON-NLS-1$
            if (transaction.length == BceeWebSiteTransactionsImpl.CSV_FIELDS) {
                if (transaction[0].equals("Transaktionsdatum")) { //$NON-NLS-1$
                    continue;
                }
                final Umsatz umsatz = this.loadLine(transaction);
                transactions.add(umsatz);

            }
        }

    }

    /**
     * L&auml;d eine Zeile aus der CSV-Datei in einen Umsatz.
     *
     * @param transaction
     *            die Umsatzzeile aus dem CSV
     * @return den Umsatz
     * @throws RemoteException
     */
    private Umsatz loadLine(final String[] transaction) throws RemoteException { // NOPMD

        // Datum,Beschreibung,Valuta,Betrag in EUR,
        // ,Auszug,Saldo,Transaktion,Mitteilung1,Mitteilung2,Mitteilung3,Mitteilung4

        final HBCIDBService dbService = Settings.getDBService();
        final SimpleDateFormat sdf = new SimpleDateFormat("dd/MM/yyyy", //$NON-NLS-1$
                Locale.getDefault());
        final Umsatz umsatz = dbService.createObject(Umsatz.class, null);

        // Buchungsdatum
        Date date = null;
        try {
            date = sdf.parse(transaction[0]);
        } catch (final ParseException e) {
            Logger.error(I18N.tr("BceeWebSiteTransactionsImpl.parseBookingDateFailed", //$NON-NLS-1$
                    e.getMessage()), e);
        }
        umsatz.setDatum(date);

        // Zweck
        final String zweck = transaction[1] + "\n" + transaction[7] + "\n" //$NON-NLS-1$ //$NON-NLS-2$
                + transaction[8] + "\n" + transaction[9] + "\n" //$NON-NLS-1$//$NON-NLS-2$
                + transaction[10];
        Utils.loadUmsatz(umsatz, zweck);

        // Valuta
        try {
            date = sdf.parse(transaction[2]);
        } catch (final ParseException e) {
            Logger.error(I18N.tr("BceeWebSiteTransactionsImpl.parseValutaDateFailed", //$NON-NLS-1$
                    e.getMessage()), e);
        }
        umsatz.setValuta(date);

        // Betrag in EUR
        umsatz.setBetrag(Utils.convertStringToDouble(transaction[3]));

        // Auszug
        if (StringUtils.isNotEmpty(transaction[4])) {
            umsatz.setKommentar(I18N.tr("BceeWebSiteTransactionsImpl.accountSheet") + transaction[4]); //$NON-NLS-1$
        }

        // Saldo
        if (StringUtils.isNotEmpty(transaction[5])) {
            umsatz.setSaldo(Utils.convertStringToDouble(transaction[5]));
        }

        // Transaktion
        umsatz.setArt(transaction[6]);

        return umsatz;
    }
}
