/*******************************************************************************
 * Copyright 2015 - 2020 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.mahoe.banks.bos.web;

import java.rmi.RemoteException;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.List;
import java.util.Locale;

import org.openqa.selenium.By;
import org.openqa.selenium.WebDriver;
import org.openqa.selenium.WebElement;
import org.openqa.selenium.support.ui.ExpectedConditions;
import org.openqa.selenium.support.ui.WebDriverWait;

import de.frankmuenster.mahoe.banks.bos.BankOfScotland;
import de.frankmuenster.mahoe.statements.AbstractAccountStatementFinder;
import de.frankmuenster.mahoe.statements.MahoeAccountStatement;
import de.frankmuenster.mahoe.utils.I18N;
import de.frankmuenster.mahoe.utils.WebUtils;
import de.frankmuenster.mahoe.web.WebSiteException;
import de.willuhn.logging.Logger;

/**
 * Class to get account statements from the banks web page
 *
 * @author Frank Mnster
 */
public class BosAccountStatementFinder extends AbstractAccountStatementFinder<WebDriver, BosWebAccountInformation> {

    /**
     * Creates a new BosAccountStatementFinder object with a webClient and
     * account information.
     *
     * @param accountInfo
     *            the {@link BosWebAccountInformation} object
     * @param driver
     *            the {@link BosWebSite} object
     */
    public BosAccountStatementFinder(final BosWebAccountInformation accountInfo, final WebDriver driver) {

        super(driver, accountInfo);
    }

    @Override
    protected void startSearch() throws WebSiteException {

        final SimpleDateFormat sdf = new SimpleDateFormat("dd.MM.yyyy", Locale.GERMAN); //$NON-NLS-1$

        String accountNumber;
        try {
            accountNumber = this.accountInfo.getKonto().getKontonummer();
            BosWebSiteHelper.findAndOpenAccount(this.webClient, accountNumber);

            this.openAccountSheets();

            // Loop over table
            // Search table
            final WebElement table = this.webClient.findElement(By.xpath(
                    "/html/body/form/table/tbody/tr/td/div[2]/table/tbody/tr[2]/td/div/div/span[2]/div[2]/table")); //$NON-NLS-1$

            Logger.debug("Table: " + table.getText()); //$NON-NLS-1$

            // and now get and loop over rows
            final List<WebElement> rows = table.findElements(By.tagName("tr")); //$NON-NLS-1$

            for (final WebElement row : rows) {
                Logger.debug("Zellen: " + row.findElements(By.tagName(WebUtils.HTML_TAG_TD)).size()); //$NON-NLS-1$
                final WebElement cell1 = row.findElement(By.xpath("td[1] | th[1]")); //$NON-NLS-1$
                if ("Auszugsnummer".equals(cell1.getText())) { //$NON-NLS-1$
                    continue;
                }
                final WebElement link = row.findElement(By.xpath("td[1]/a")); //$NON-NLS-1$
                final String toDate = row.findElement(By.xpath("td[3]")).getText(); //$NON-NLS-1$
                final Date pdfDate = sdf.parse(toDate);

                if (pdfDate.before(this.lastStatementDate)) {
                    // OK, here we have finished
                    break; // NOPMD
                }
                // Get year and number
                final String yearText = cell1.getText().substring(0, 4);
                final String numberText = cell1.getText().substring(cell1.getText().indexOf('-') + 1);

                final Integer year = Integer.valueOf(yearText);
                final Integer number = Integer.valueOf(numberText);

                // get the expected file name
                String href = link.getAttribute(WebUtils.HTML_TAG_HREF);
                href = href.substring(29, href.length() - 2);
                final String[] args = href.split(","); //$NON-NLS-1$
                final String expectedFileName = args[2].substring(1, args[2].length() - 1);

                // Download PDF as byte array
                final byte[] data = WebUtils.getSeleniumPdfAsByteArray(this.webClient, link, expectedFileName);
                if (data != null) {
                    // create an MahoeAccountStatement
                    final MahoeAccountStatement statementInfo = new MahoeAccountStatement(pdfDate, BankOfScotland.NAME); // NOPMD
                    statementInfo.setYear(year);
                    statementInfo.setNumber(number);
                    // statement to internal list
                    this.addStatement(statementInfo, data);
                }

            }

        } catch (final RemoteException e) {
            Logger.error(I18N.tr("BosAccountStatementFinder.accountNumberNotSet")); //$NON-NLS-1$
            throw new WebSiteException(I18N.tr("BosAccountStatementFinder.accountNumberNotSet"), e); //$NON-NLS-1$
        } catch (final ParseException e) {
            Logger.error(I18N.tr("TargoBankWebSite.parsePdfDateFailed", e.getMessage()), e); //$NON-NLS-1$
        }

    }

    /**
     * Opens the account sheets page
     */
    private void openAccountSheets() {

        final WebElement actSheetBtn = this.webClient.findElement(
                By.xpath("/html/body/form/table/tbody/tr/td/div[2]/table/tbody/tr[2]/td/div/div/div/ul/li[1]/input")); //$NON-NLS-1$
        actSheetBtn.click();

        final WebDriverWait wait = new WebDriverWait(this.webClient, 10);
        wait.until(ExpectedConditions.presenceOfElementLocated(By
                .xpath("/html/body/form/table/tbody/tr/td/div[2]/table/tbody/tr[2]/td/div/div/span[2]/div[2]/table"))); //$NON-NLS-1$

    }

}
