/*******************************************************************************
 * Copyright 2015 - 2020 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.mahoe.banks.santander.clearing;

import java.math.BigDecimal;
import java.math.RoundingMode;
import java.rmi.RemoteException;
import java.util.Date;

import org.apache.commons.lang.NullArgumentException;
import org.apache.commons.lang.StringUtils;

import de.frankmuenster.mahoe.MahoeSynchronizeBackend;
import de.frankmuenster.mahoe.utils.I18N;
import de.frankmuenster.mahoe.utils.Utils;
import de.willuhn.datasource.rmi.DBIterator;
import de.willuhn.jameica.hbci.Settings;
import de.willuhn.jameica.hbci.rmi.AuslandsUeberweisung;
import de.willuhn.jameica.hbci.rmi.HBCIDBService;
import de.willuhn.jameica.hbci.rmi.Konto;
import de.willuhn.jameica.hbci.rmi.Transfer;
import de.willuhn.jameica.hbci.rmi.Umsatz;
import de.willuhn.jameica.messaging.StatusBarMessage;
import de.willuhn.jameica.system.Application;
import de.willuhn.logging.Logger;
import de.willuhn.util.ApplicationException;

/**
 * Bean mit Ausgleichbuchungsinformationen
 *
 * @author Frank Mnster
 */
public class ClearingEntry {

    /**
     * Der Ausgleichsbetrag
     */
    private final BigDecimal betrag;

    /**
     * Das F&auml;lligkitsdatum
     */
    private final Date datum;

    /**
     * Die Kreditkartennummer
     */
    private final String kartenNummer;

    /**
     * Der Karteninhaber
     */
    private final String inhaber;

    /**
     * Die IBAN
     */
    private final String iban;

    /**
     * Die BIC
     */
    private final String bic;

    /**
     * Das Abbuchungskonto
     */
    private Konto konto;

    /**
     * Erstellt eine neue Ausgleichsbuchung mit den angegebenen Daten.
     *
     * @param kartenNumber
     * @param inhaber
     * @param iban
     * @param bic
     * @param datum
     * @param betrag
     * @throws NullArgumentException
     */
    public ClearingEntry(final String kartenNumber, final String inhaber, final String iban, final String bic,
            final Date datum, final BigDecimal betrag) throws NullArgumentException {

        this.kartenNummer = kartenNumber;
        this.inhaber = inhaber;
        this.iban = iban;
        this.bic = bic;
        this.betrag = betrag.setScale(2, RoundingMode.HALF_UP);
        this.datum = datum;
        // Alle Felder gesetzt
        this.checkParameter();
    }

    /**
     * berprpt, ob eine Ausgleichsbuchung in den letzten 365 Tagen mit den
     * selben Betrag schon existiert. Es werden alle Hibiscus-Konten durchsucht.
     *
     * @return <i>true</i>, wenn die Ausgleichbuchung schon existiert,
     *         <i>false</i> sonst.
     */
    public boolean exists() {

        boolean exists = false;

        try {
            final HBCIDBService hcbiService = Settings.getDBService();

            final DBIterator<Konto> accounts = hcbiService.createList(Konto.class);

            if (this.konto != null) {
                final String ktoID = this.konto.getID();
                accounts.addFilter("(id <> " + ktoID + ")"); //$NON-NLS-1$ //$NON-NLS-2$
            }

            while (accounts.hasNext() && !exists) {
                final Konto account = accounts.next();

                @SuppressWarnings("unchecked")
                final DBIterator<Umsatz> transactions = account.getUmsaetze(35);
                while (transactions.hasNext()) {
                    final Umsatz umsatz = transactions.next();
                    // check if equal
                    if (equalTransfer(umsatz)) {
                        // set exists to true and exit
                        exists = true;
                        break;
                    }
                }
            }

            // date 40 days back
            final Date limit = Utils.nowMinusDays(41);

            final DBIterator<AuslandsUeberweisung> transfers = hcbiService.createList(AuslandsUeberweisung.class);
            transfers.setOrder("ORDER BY termin desc"); //$NON-NLS-1$

            // Check transfers until either the date is NULL, before the limit
            // or an existing transfer is found
            boolean dateReached = false;
            while (transfers.hasNext() && !exists && !dateReached) {

                // get next transfer
                final AuslandsUeberweisung transfer = transfers.next();
                // check date
                final Date date = transfer.getAusfuehrungsdatum();
                if (date != null && date.before(limit)) {
                    dateReached = true;
                } else {
                    // check if equal
                    if (equalTransfer(transfer)) {
                        // set exists to true and exit
                        exists = true;
                    }
                }
            }
            Logger.debug("Exists: " + exists); //$NON-NLS-1$

        } catch (final RemoteException e) {
            // Auf true setzen, damit die berweisung nicht ggf. doppelt
            // angelegt wird.
            exists = true;
            final String msg = I18N.tr("ClearingEntry.searchClearingTransactionFailed"); //$NON-NLS-1$
            Logger.error(msg, e);
            final StatusBarMessage msgUser = new StatusBarMessage(msg, StatusBarMessage.TYPE_ERROR);
            Application.getMessagingFactory().sendMessage(msgUser);
        }
        return exists;
    }

    /**
     * Erstellt die &Uuml;berweisung in Hibiscus. Passiert schon mal nur, wenn
     * der angegeben Betrag kleiner 0 ist.
     *
     * @return <i>true</i> falls die &Uuml;berweisung angelegt werden konnte,
     *         <i>false</i> sonst und wenn keine Expection ausgel&ouml;st wurde.
     * @throws NullArgumentException
     */
    public boolean saveEntry() throws NullArgumentException {

        if (this.konto == null) {
            throw new NullArgumentException(I18N.tr("ClearingEntry.clearingAccountIsNull")); //$NON-NLS-1$
        }
        boolean saved = false;
        if (this.betrag.signum() < 0) {
            try {
                final AuslandsUeberweisung clearingTransac = (AuslandsUeberweisung) Settings.getDBService()
                        .createObject(AuslandsUeberweisung.class, null);
                clearingTransac.setKonto(this.konto);
                // Umlaute ersetzten
                clearingTransac.setGegenkontoName(Utils.replaceUmlauts(this.inhaber));
                clearingTransac.setGegenkontoNummer(this.iban);
                clearingTransac.setGegenkontoBLZ(this.bic);
                final Double myBetrag = this.betrag.negate().doubleValue();
                if (!myBetrag.isNaN()) {
                    clearingTransac.setBetrag(myBetrag);
                }
                clearingTransac.setZweck(this.kartenNummer);
                // setzte target date if not NULL
                if (this.datum != null) {
                    clearingTransac.setTermin(this.datum);
                }

                // Speichern
                clearingTransac.store();
                saved = true;
            } catch (final RemoteException e) {
                Logger.error(I18N.tr("ClearingEntry.createSepaTransactionFailed") + e.getMessage(), e); //$NON-NLS-1$
            } catch (final ApplicationException e) {
                Logger.error(I18N.tr("ClearingEntry.saveSepaTransactionFailed") + e.getMessage(), e); //$NON-NLS-1$
            }
        }
        return saved;
    }

    /**
     * Setzt das Abbuchungskonto
     *
     * @param konto
     *            das Abbuchungskonto
     * @throws NullArgumentException
     * @throws RemoteException
     */
    public void setKonto(final Konto konto) throws NullArgumentException, RemoteException {

        if (konto == null) {
            throw new NullArgumentException(I18N.tr("ClearingEntry.clearingAccountIsNull")); //$NON-NLS-1$
        }
        final String ktoId = konto.getMeta(MahoeSynchronizeBackend.CLEARING_ACCOUNT, StringUtils.EMPTY);
        if (StringUtils.isNotEmpty(ktoId)) {
            this.konto = Utils.getKonto(ktoId);
        }
    }

    /**
     * &Uuml;berpr&uuml;pt die Parameter die zum Erstellen der Klasse angegeben
     * wurden.
     *
     * @throws NullArgumentException
     */
    private void checkParameter() throws NullArgumentException {

        if (this.kartenNummer == null) {
            throw new NullArgumentException(I18N.tr("ClearingEntry.cardNumberIsNUll")); //$NON-NLS-1$
        }
        if (this.inhaber == null) {
            throw new NullArgumentException(I18N.tr("ClearingEntry.ownerIsNull")); //$NON-NLS-1$
        }
        if (this.iban == null) {
            throw new NullArgumentException(I18N.tr("ClearingEntry.ibanIsNull")); //$NON-NLS-1$
        }
        if (this.bic == null) {
            throw new NullArgumentException(I18N.tr("ClearingEntry.BicIsNull")); //$NON-NLS-1$
        }
        if (this.betrag == null) {
            throw new NullArgumentException(I18N.tr("ClearingEntry.amountIsNull")); //$NON-NLS-1$
        }
    }

    /**
     * Checks if the given Transfer object is equal to this
     * {@linkplain ClearingEntry}
     *
     * @param transfer
     *            the {@linkplain Transfer} object to check against
     * @return <b>true</b> in case the Transfer object represents this
     *         ClearingEntry, <b>false</b> else
     * @throws RemoteException
     *             in case some attributes of the Transfer object could not be
     *             read from the database
     */
    private boolean equalTransfer(final Transfer transfer) throws RemoteException {

        // Initialize equal variable
        boolean equal = false;
        // Only if transfer object not NULL
        if (transfer != null) {
            // get the receiver name
            final String gkInhaber = StringUtils.isNotEmpty(transfer.getGegenkontoName()) ? transfer.getGegenkontoName()
                    : StringUtils.EMPTY;
            // get the amount
            final BigDecimal transBetrag = Utils.roundDoubleToBigDecimal(transfer.getBetrag(), 2);
            // get card number
            final String transKartenNummer = StringUtils.isNotEmpty(transfer.getZweck()) ? transfer.getZweck()
                    : StringUtils.EMPTY;
            // and compare with this clearance entry
            final String inhaber2 = Utils.replaceUmlauts(this.inhaber);
            if (inhaber2.equalsIgnoreCase(gkInhaber) && this.kartenNummer.equals(transKartenNummer)
                    && this.betrag.abs().compareTo(transBetrag) == 0) {
                equal = true;
            }
        }
        // return result
        return equal;

    }

}
