/*******************************************************************************
 * Copyright 2015 - 2020 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.mahoe.utils;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.nio.file.Files;
import java.util.Date;
import java.util.StringTokenizer;

import org.openqa.selenium.Cookie;
import org.openqa.selenium.WebDriver;
import org.openqa.selenium.WebDriverException;

import de.frankmuenster.mahoe.MahoeMain;
import de.willuhn.jameica.messaging.StatusBarMessage;
import de.willuhn.jameica.system.Application;
import de.willuhn.logging.Logger;
import de.willuhn.util.ApplicationException;

/**
 * Utility class to handles cookies. Offers methods to save and restore cookies
 * for a web domain
 * 
 * @author Frank Mnster
 * @since 2.4.0
 *
 */
public final class SeleniumCookie {

    /**
     * Number of cookie parameters
     */
    private static final int COOKIE_PARAMS_COUNT = 7;

    /**
     * I18N translation string for cookie backup error
     */
    private static final String I18N_COOKIE_BACKUP_ERROR = "SeleniumCookie.backupError"; //$NON-NLS-1$

    /**
     * Adds a {@link Cookie} to a {@link WebDriver}
     *
     * @param driver
     *            the {@link WebDriver}
     * @param cookie
     *            the {@link Cookie} to add
     */
    private static void addCookie(final WebDriver driver, final Cookie cookie) {
        try {
            // Stored cookie to your current session
            driver.manage().addCookie(cookie);
        } catch (final IllegalArgumentException | WebDriverException e) {
            // Only log this as it can happen
            Logger.warn(I18N.tr("SeleniumCookie.cookieRecovverErrorLog", cookie.getName()) //$NON-NLS-1$
                    + e.getLocalizedMessage());
        }

    }

    /**
     * Reads the cookies from the backup file
     *
     * @param driver
     *            Selenium {@link WebDriver} to add cookies to
     * @param backupCookieFile
     *            the backup file name for this bank
     * @throws ApplicationExceptionin
     *             case the cookies could not be restored
     */
    public static void read(final WebDriver driver, final String backupCookieFile) {

        // path to plugin
        final String pluginPath = Application.getPluginLoader().getPlugin(MahoeMain.class).getResources().getWorkPath();

        // create file named Cookies to store Login Information
        final File file = new File(pluginPath, backupCookieFile);

        try (FileReader fileReader = new FileReader(file); BufferedReader bufReader = new BufferedReader(fileReader)) {
            String strline;
            // loop over lines. Each lien is a cookie
            while ((strline = bufReader.readLine()) != null) {
                // separate cookie sections
                final StringTokenizer token = new StringTokenizer(strline, ";"); //$NON-NLS-1$ //NOPMD

                // Only read cookies with 7 parameters
                if (token.countTokens() == SeleniumCookie.COOKIE_PARAMS_COUNT) {
                    final String name = token.nextToken();
                    final String value = token.nextToken();
                    final String domain = token.nextToken();
                    final String path = token.nextToken();
                    Date expiry = null;

                    final String val = token.nextToken();
                    if (!"null".equals(val)) { //$NON-NLS-1$ //NOPMD
                        expiry = new Date(Long.valueOf(val)); // NOPMD
                    }
                    final Boolean isSecure = Boolean.parseBoolean(token.nextToken());
                    final Boolean isHttpOnly = Boolean.parseBoolean(token.nextToken());

                    final Cookie ck = new Cookie(name, value, domain, path, expiry, isSecure, isHttpOnly); // NOPMD

                    Logger.debug("Cookies: " + ck); //$NON-NLS-1$

                    SeleniumCookie.addCookie(driver, ck);

                }
            }

        } catch (final FileNotFoundException e) {
            // Only log a message
            Logger.info(I18N.tr("SeleniumCookie.backupFileNotFoundLogMsg") + e.getLocalizedMessage()); //$NON-NLS-1$
        } catch (final IOException e) {
            // Display as warning
            Logger.warn("Cookies Wiederherstellung nicht mglich. Fehler: " + e.getLocalizedMessage()); //$NON-NLS-1$
            final StatusBarMessage msg = new StatusBarMessage(
                    "Browser-Cookies konnten nicht wieder hergstellt werden. Evtl. mussen Logindaten fter eingegeben werden.", //$NON-NLS-1$
                    StatusBarMessage.TYPE_INFO);
            Application.getMessagingFactory().sendMessage(msg);
        }

    }

    /**
     * Writes the cookies in the backup file
     *
     * @param driver
     *            Seleniuem {@link WebDriver} to get the cookies from
     * @param backupCookieFile
     *            the backup file name for this bank
     * @throws ApplicationException
     *             in case the cookies could not be stored
     */
    public static void write(final WebDriver driver, final String backupCookieFile) throws ApplicationException {

        // path to plugin
        final String path = Application.getPluginLoader().getPlugin(MahoeMain.class).getResources().getWorkPath();

        // create file named Cookies to store Login Information
        final File file = new File(path, backupCookieFile);
        try {
            // Delete old file if exists
            Files.deleteIfExists(file.toPath());
            if (!file.createNewFile()) {
                throw new ApplicationException(I18N.tr(SeleniumCookie.I18N_COOKIE_BACKUP_ERROR));
            }
        } catch (final IOException e) {
            Logger.error(I18N.tr(SeleniumCookie.I18N_COOKIE_BACKUP_ERROR), e);
            throw new ApplicationException(I18N.tr(SeleniumCookie.I18N_COOKIE_BACKUP_ERROR), e);
        }

        try (FileWriter fileWrite = new FileWriter(file); BufferedWriter bufWwrite = new BufferedWriter(fileWrite)) {
            // loop for getting the cookie information

            // Use StringBuffer as it is quicker
            final StringBuffer cookieBuf = new StringBuffer(50);

            for (final Cookie cookie : driver.manage().getCookies()) {

                // get Date as Long string or null
                String date = "null"; //$NON-NLS-1$
                if (cookie.getExpiry() != null) {
                    date = String.valueOf(cookie.getExpiry().getTime());
                }
                // clear StringBuffer
                cookieBuf.setLength(0);
                cookieBuf.append(cookie.getName()).append(';').append(cookie.getValue()).append(';')
                        .append(cookie.getDomain()).append(';').append(cookie.getPath()).append(';').append(date)
                        .append(';').append(cookie.isSecure()).append(';').append(cookie.isHttpOnly());
                bufWwrite.write(cookieBuf.toString());
                bufWwrite.newLine();
            }

        } catch (final IOException e) {
            Logger.error(I18N.tr(SeleniumCookie.I18N_COOKIE_BACKUP_ERROR), e);
            throw new ApplicationException(I18N.tr(SeleniumCookie.I18N_COOKIE_BACKUP_ERROR), e);
        }

    }

    /**
     * Private calls constructor
     */
    private SeleniumCookie() {
        // Utility class --> do not instantiate
    }

}
