/*******************************************************************************
 * Copyright 2015 - 2020 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.mahoe.utils;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.nio.file.Files;
import java.nio.file.Path;
import java.time.Duration;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.apache.commons.io.IOUtils;
import org.apache.http.entity.ContentType;
import org.openqa.selenium.JavascriptExecutor;
import org.openqa.selenium.TimeoutException;
import org.openqa.selenium.WebDriver;
import org.openqa.selenium.WebElement;
import org.openqa.selenium.support.ui.ExpectedConditions;
import org.openqa.selenium.support.ui.FluentWait;
import org.openqa.selenium.support.ui.WebDriverWait;

import com.gargoylesoftware.htmlunit.WebResponse;
import com.gargoylesoftware.htmlunit.html.HtmlAnchor;

import de.frankmuenster.mahoe.web.AbstractSeleniumWebSite;
import de.frankmuenster.mahoe.web.WebSiteException;
import de.frankmuenster.mahoe.web.WebSiteExceptionTypes;
import de.willuhn.logging.Logger;
import de.willuhn.util.ApplicationException;

/**
 * Utility class for web methods
 *
 * @author Frank Mnster
 */
public final class WebUtils {

    /**
     * The maximum number of new files during a download in the download folder.
     */
    private static final int MAX_NUMBER_NEW_FILES = 1;

    /**
     * HTML tag href
     */
    public static final String HTML_TAG_HREF = "href"; //$NON-NLS-1$

    /**
     * HTML tag "td"
     */
    public static final String HTML_TAG_TD = "td"; //$NON-NLS-1$

    /**
     * HTML tag for the first cell "td[1]"
     */
    public static final String HTML_TAG_TD1 = WebUtils.HTML_TAG_TD + "[1]"; //$NON-NLS-1$

    /**
     * HTML tag "tr"
     */
    public static final String HTML_TAG_TR = "tr"; //$NON-NLS-1$

    /**
     * HTML tag "table"
     */
    public static final String HTML_TAG_TABLE = "table"; //$NON-NLS-1$

    /**
     * HTML tag for table body "tbody"
     */
    public static final String HTML_TAG_TBODY = "tbody"; //$NON-NLS-1$

    /**
     * HTML tag anchor "a"
     */
    public static final String HTML_TAG_ANCHOR = "a"; //$NON-NLS-1$

    /**
     * HTML tag paragraph "p"
     */
    public static final String HTML_TAG_PARAGRAPH = "p"; //$NON-NLS-1$

    /**
     * Read the account statement PDF from an HTMLUnit anchor into a byte array.
     *
     * @param link
     *            the HMTLUnit anchor to the PDF
     * @return the PDF as byte array
     */
    public static byte[] getPdfAsByteArray(final HtmlAnchor link) {

        byte[] data = null;
        try (InputStream inputStream = link.click().getWebResponse().getContentAsStream()) {

            data = IOUtils.toByteArray(inputStream);

        } catch (final IOException exception) {
            Logger.error(exception.getMessage());
        }

        return data;

    }

    /**
     * Returns the query parameters of an URI
     *
     * @param query
     * @return a map with all param value pairs
     */
    public static Map<String, String> getQueryMap(final URI uri) {

        final Map<String, String> map = new ConcurrentHashMap<>();
        final String query = uri.getQuery();
        if (query != null) {
            final String[] params = query.split("&"); //$NON-NLS-1$
            for (final String param : params) {
                final String name = param.split("=")[0]; //$NON-NLS-1$
                final String value = param.split("=")[1]; //$NON-NLS-1$
                map.put(name, value);
            }
        }
        return map;
    }

    /**
     * Read the account statement PDF into a byte array.
     *
     * @param driver
     *            the {@link WebDriver} to use
     * @param linkElement
     *            the link on the page to the PDF
     * @param expectedFileName
     * @return the PDF as byte array
     */
    public static byte[] getSeleniumPdfAsByteArray(final WebDriver driver, final WebElement linkElement)
            throws WebSiteException {

        byte[] data = null;

        Logger.debug("Downloading file with unknown name"); //$NON-NLS-1$

        try {

            final File dir = AbstractSeleniumWebSite.getDownloadsDirectory().toFile();

            // get the actual files in the download folder
            final HashSet<File> existingFileList = new HashSet<>(Arrays.asList(dir.listFiles()));
            // how many have we
            final int fileCount = existingFileList.size();

            // scroll into view
            ((JavascriptExecutor) driver).executeScript("arguments[0].scrollIntoView(true); window.scrollBy(0, -350);", //$NON-NLS-1$
                    linkElement);
            final WebDriverWait wait1 = new WebDriverWait(driver, 10);
            wait1.until(ExpectedConditions.elementToBeClickable(linkElement));

            // trigger the download
            linkElement.click();
            // wait until we have more files in the download folder
            final FluentWait<WebDriver> wait = new FluentWait<>(driver).withTimeout(Duration.ofSeconds(60))
                    .pollingEvery(Duration.ofMillis(200));

            // For the file system to be ready
            Utils.waitIgnoreThreadInterrupts(1000);

            Logger.debug("Files in download dir: " + dir.list().length); //$NON-NLS-1$
            wait.until((final WebDriver wd) -> dir.list().length > fileCount);

            // get new file sin folder
            final HashSet<File> newFileList = new HashSet<>(Arrays.asList(dir.listFiles()));
            // remove the existing once
            newFileList.removeAll(existingFileList);
            // check if one 1 new file
            if (newFileList.size() == WebUtils.MAX_NUMBER_NEW_FILES) {
                // get the file
                final File downloadedFile = newFileList.iterator().next();
                Logger.info(I18N.tr("WebUtils.downloadedFileName") + downloadedFile.getAbsolutePath()); //$NON-NLS-1$
                // read file in memory
                data = Files.readAllBytes(downloadedFile.toPath());
                // delete file again
                Files.delete(downloadedFile.toPath());
            } else {
                Logger.error(I18N.tr("WebUtils.moreThanOneNewFileInDownloadFolderMsg")); //$NON-NLS-1$
                throw new WebSiteException(I18N.tr("WebUtils.moreThanOneNewFileInDownloadFolderExp"), //$NON-NLS-1$
                        WebSiteExceptionTypes.TRANSFER_FAILED);
            }

        } catch (final IOException | TimeoutException | ApplicationException e) {
            Logger.error(I18N.tr("WebUtils.errorDownloadAccountSheet1"), e); //$NON-NLS-1$
            throw new WebSiteException(I18N.tr("WebUtils.errorDownloadAccountSheet1"), //$NON-NLS-1$
                    WebSiteExceptionTypes.TRANSFER_FAILED, e);
        }

        return data;

    }

    /**
     * Read the account statement PDF into a byte array.
     *
     * @param driver
     *            the {@link WebDriver} to use
     * @param linkElement
     *            the link on the page to the PDF
     * @param expectedFileName
     * @return the PDF as byte array
     */
    public static byte[] getSeleniumPdfAsByteArray(final WebDriver driver, final WebElement linkElement,
            final String expectedFileName) {

        byte[] data = null;

        Logger.debug("Downloading file: " + expectedFileName); //$NON-NLS-1$

        linkElement.click();
        try {

            final FluentWait<WebDriver> wait = new FluentWait<>(driver).withTimeout(Duration.ofSeconds(60))
                    .pollingEvery(Duration.ofMillis(200));
            final File fileToCheck = AbstractSeleniumWebSite.getDownloadsDirectory().resolve(expectedFileName).toFile();

            final Path dir = AbstractSeleniumWebSite.getDownloadsDirectory();

            dir.getNameCount();

            // Wait until download is complete
            wait.until((final WebDriver wd) -> fileToCheck.exists());
            // read file in memory
            data = Files.readAllBytes(fileToCheck.toPath());
            // delete file again
            Files.delete(fileToCheck.toPath());

        } catch (final ApplicationException | IOException | TimeoutException e) {
            Logger.error(I18N.tr("WebUtils.errorDownloadAccountSheet", expectedFileName), e); //$NON-NLS-1$
        }

        return data;

    }

    /**
     * Scroll to an element so it is visible
     */
    public static void scrollTo(final WebDriver driver, final WebElement element) {

        // Scroll and click on list element
        final JavascriptExecutor executor = (JavascriptExecutor) driver;
        executor.executeScript("arguments[0].scrollIntoView(false);", //$NON-NLS-1$
                element);

    }

    /**
     * berprft die Webantwort, ob der http Statuscode 200 ist und die Antwort
     * im JSON-Format vorliegt.
     *
     * @param webResponse
     * @return true wenn HHTP code 200 und Antwortinhalt application/json ist,
     *         sonst false.
     */
    public static boolean webResponseIsJson(final WebResponse webResponse) { // NOPMD

        boolean answerOK = false;
        if (webResponse.getStatusCode() == org.apache.http.HttpStatus.SC_OK
                && ContentType.APPLICATION_JSON.getMimeType().equalsIgnoreCase(webResponse.getContentType())) { // $NON-NLS-1$
            answerOK = true;

        } else {
            // Ein Debuginfos loggen
            Logger.debug("JSON-Check-Status: " + webResponse.getStatusCode()); //$NON-NLS-1$
            Logger.debug("JSON-Check-MimeType: " + webResponse.getContentType()); //$NON-NLS-1$
            Logger.debug("JSON-Check-Inhalt: " + webResponse.getContentAsString()); //$NON-NLS-1$
        }
        return answerOK;
    }

    /**
     * private class constructor
     */
    private WebUtils() {

        // intentionally left blank
    }

}
