/*******************************************************************************
 * Copyright 2015 - 2020 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.mahoe.web;

import java.lang.reflect.GenericDeclaration;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.lang.reflect.TypeVariable;
import java.util.Calendar;
import java.util.Date;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import de.frankmuenster.mahoe.messages.AbstractBankMessages;
import de.frankmuenster.mahoe.utils.I18N;
import de.frankmuenster.mahoe.utils.Utils;
import de.willuhn.jameica.hbci.rmi.Konto;
import de.willuhn.logging.Logger;
import de.willuhn.util.ApplicationException;

/**
 * Abstrakte WebSite Klasse.
 *
 * @author Frank Mnster
 * @param <C>
 *            the client class, HTMLUnit or Selemium
 * @param <I>
 * @param <M>
 */
public abstract class AbstractWebSite<C, I extends WebAccountInformation, M extends AbstractBankMessages<C>>
        implements WebSite<I> {

    /**
     * HTML Tag einer Tabelle
     */
    protected static final String TABLE_TAG = "table"; //$NON-NLS-1$

    /**
     * Non breaking space unicode
     */
    public static final String NON_BREAK_SPACE = "\u00a0"; //$NON-NLS-1$

    /**
     * Unser WebClient
     */
    protected C webClient;

    /**
     * Die Accountinfos
     */
    protected I accountInfo;

    /**
     * Searches and returns the class of a generic Type with the given position
     * in the given superclass. Takes into account the complete hierarchy.
     *
     * @param parameterizedSubClass
     * @param genericSuperClass
     * @param pos
     * @return the class
     */
    @SuppressWarnings("unchecked")
    public static <T> Class<T> getGenericClassParameter(final Class<?> parameterizedSubClass,
            final Class<?> genericSuperClass, final int pos) {
        // a mapping from type variables to actual values (classes)
        final Map<TypeVariable<?>, Class<?>> mapping = new ConcurrentHashMap<>();

        Class<?> klass = parameterizedSubClass;
        while (klass != null) {
            final Type type = klass.getGenericSuperclass();
            if (type instanceof ParameterizedType) {
                final ParameterizedType parType = (ParameterizedType) type;
                final Type rawType = parType.getRawType();
                if (rawType == genericSuperClass) { // NOPMD
                    // found
                    final Type typ = parType.getActualTypeArguments()[pos];
                    if (typ instanceof Class<?>) {
                        return (Class<T>) typ; // NOPMD
                    } else {
                        return (Class<T>) mapping.get(typ);
                    }
                }
                // resolve
                final Type[] vars = ((GenericDeclaration) parType.getRawType()).getTypeParameters();
                final Type[] args = parType.getActualTypeArguments();
                for (int i = 0; i < vars.length; i++) {
                    if (args[i] instanceof Class<?>) {
                        mapping.put((TypeVariable<?>) vars[i], (Class<?>) args[i]);
                    } else {
                        mapping.put((TypeVariable<?>) vars[i], mapping.get(args[i]));
                    }
                }
                klass = (Class<?>) rawType;
            } else {
                klass = klass.getSuperclass();
            }
        }
        throw new IllegalArgumentException(
                I18N.tr("AbstractWebSite.noSuperGenericSuperType", parameterizedSubClass, genericSuperClass)); //$NON-NLS-1$
    }

    /**
     * Unser Konstruktor. Nur zugnglich fr Sub-Klassen.
     *
     * @param konto
     *            das Konto
     * @throws WebSiteException
     */
    protected AbstractWebSite(final Konto konto) throws WebSiteException {

        java.util.logging.Logger.getLogger("com.gargoylesoftware.htmlunit") //$NON-NLS-1$
                .setLevel(java.util.logging.Level.OFF);
        this.webClient = this.getWebClient();

        this.setAccountBasicInfo(konto);

    }

    /**
     * Ldt neue Bank-Nachrichten.
     *
     * @param accountInfo
     * @throws WebSiteException
     * @throws ApplicationException
     */

    @Override
    public int getMessages(final I accountInfo) throws WebSiteException, ApplicationException {

        int saveMsg;

        // get the BankMessages class
        final Class<M> bankMessageClazz = AbstractWebSite.getGenericClassParameter(this.getClass(),
                AbstractWebSite.class, 2);

        // now get webclient class (HTMLUnit or Selenium)
        final Class<
                C> webClientClass = AbstractWebSite.getGenericClassParameter(this.getClass(), AbstractWebSite.class, 0);

        try {
            final M messages = bankMessageClazz.getDeclaredConstructor(webClientClass).newInstance(this.webClient);

            // search for last message from this bank account
            final Date lastMsgDate = messages.getLastMsgDate(accountInfo);
            final Date lastTransactionDate = accountInfo.getLastTransactionDate();
            final Date startDate = lastMsgDate.before(lastTransactionDate) ? lastMsgDate : lastTransactionDate;
            messages.readMessages(startDate);
            saveMsg = messages.saveMessage();

        } catch (final ApplicationException e) {
            Logger.error(I18N.tr("AbstractWebSite.getBankMessagesFailed"), //$NON-NLS-1$
                    e);
            throw e;
        } catch (InstantiationException | IllegalAccessException | IllegalArgumentException | InvocationTargetException
                | NoSuchMethodException | SecurityException e) {
            Logger.error(I18N.tr("AbstractWebSite.instanziateBankMessageClassFailed"), //$NON-NLS-1$
                    e);
            throw new ApplicationException(I18N.tr("AbstractWebSite.instanziateBankMessageClassFailed"), //$NON-NLS-1$
                    e);
        }

        return saveMsg;
    }

    /**
     * Das Datum, ab dem Umstze und Banknachrichten abgerufen werden sollen,
     * falls noch keine Umstze vorliegen.
     *
     * @return das StartDatum
     */
    protected abstract Calendar getStartdate();

    /**
     * Erstellt den bankspezifischen WebClient.
     *
     * @return den WebClient
     */
    protected abstract C getWebClient();

    /**
     * Creates a new {@link WebAccountInformation} object for the bank from the
     * given account. Then is sets the account and the last Transaction date.
     *
     * @param konto
     *            the account to take the information from.
     * @throws WebSiteException
     */
    protected final void setAccountBasicInfo(final Konto konto) throws WebSiteException {

        final Class<I> clazz = AbstractWebSite.getGenericClassParameter(this.getClass(), AbstractWebSite.class, 1);

        try {
            this.accountInfo = clazz.getDeclaredConstructor().newInstance();

            // Setze Konto
            this.accountInfo.setKonto(konto);

            // Setzte letztes Umsatzdatum
            final Date lastTransacDate = Utils.getLastUmsatzDate(konto, this.getStartdate());
            this.accountInfo.setLastTransactionDate(lastTransacDate);

        } catch (InstantiationException | IllegalAccessException | IllegalArgumentException | InvocationTargetException
                | NoSuchMethodException | SecurityException e) {
            Logger.error(I18N.tr("AbstractWebSite.settingAccountInformationFailed")); //$NON-NLS-1$
            throw new WebSiteException(I18N.tr("AbstractWebSite.settingAccountInformationFailed"), //$NON-NLS-1$
                    WebSiteExceptionTypes.READ_ACCOUNT_DETAILS_FAILED, e);
        }
    }

}
